import warnings
from tmt import tmt_recorder
from tmt.storage.json_db import DbManager
from tmt.history.context import context_manager, ContextManager, Configs
from tmt.history.utils import save
from tmt.exceptions import DuplicatedNameError
from tmt.utils.duplicates import *
from tests import BaseTest
import os
import pickle


returned_metrics = {'f1': 0.87, 'acc': 0.45, 'loss': 1e-4}


@tmt_recorder('test_exp', config_path='tests/test_config.json')
def decorated_fn(_):
    # do something
    save(returned_metrics, name='test_metrics')
    return returned_metrics


class TestDecorators(BaseTest):

    def test_recorder(self):
        metrics = decorated_fn(None)
        self.assertEqual(metrics, returned_metrics)
        self.assertIsInstance(context_manager.get(), ContextManager)
        entry = context_manager.get().entry
        db_man = DbManager(context_manager.get().config.json_db_path)
        db_entry = db_man.get_entry_by_id(entry.id)
        self.assertEqual(entry.to_dict(), db_entry.to_dict())
        self.assertTrue(os.path.exists(context_manager.get().snap_manager.snapshot_dest))
        self.assertTrue(os.path.exists(context_manager.get().get_save_path()))
        self.assertTrue(len(entry.metrics) > 0)

        # test results saving 
        self.assertTrue(os.path.exists(context_manager.get().get_save_path_with_name(name='test_metrics') + '.pkl'))
        with open(context_manager.get().get_save_path_with_name(name='test_metrics') + '.pkl', 'rb') as f:
            self.assertEqual(pickle.load(f), returned_metrics)
        self.assertTrue(len(entry.results) > 0)
        self.assertRaises(DuplicatedNameError, ContextManager, 'test_exp', 'tests/test_config.json')

    def test_duplicates(self):
        def test_fn(_):
            save(returned_metrics, name='test_metrics')
            return returned_metrics

        db_man = DbManager(self.conf.json_db_path)
        db_man.delete_all()
        fn = tmt_recorder('test_exp', config_path='tests/test_config.json')(test_fn)
        _ = fn(None)
        self.assertRaises(DuplicatedNameError, fn, None)
        fn = tmt_recorder('test_exp', config_path='tests/test_config.json', duplicate_strategy=DuplicateStrategy(DuplicatePolicy.AS_NEW_ENTRY))(test_fn)
        self.assertEqual(returned_metrics, fn(None))
        parent_id = context_manager.get().entry.id
        entries = db_man.get_entries_by_name("test_exp")
        self.assertEqual(len(entries), 2)
        for e in entries:
            self.assertEqual(len(e.other_runs), 0)

        # test sub entries
        fn = tmt_recorder('test_exp', config_path='tests/test_config.json', duplicate_strategy=DuplicateStrategy(DuplicatePolicy.AS_SUB_ENTRY))(test_fn)
        self.assertEqual(returned_metrics, fn(None))
        self.assertWarns(UserWarning, fn, None)
        fn = tmt_recorder('test_exp', config_path='tests/test_config.json', duplicate_strategy=DuplicateStrategy(DuplicatePolicy.AS_SUB_ENTRY, parent_id))(test_fn)
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter('always')
            self.assertEqual(returned_metrics, fn(None))
            self.assertEqual(0, len(w))
        entries = db_man.get_entries_by_name("test_exp")
        self.assertEqual(len(entries), 2)
        for e in entries:
            if e.id == parent_id:
                self.assertTrue(1 <= len(e.other_runs) <= 3)

    def test_description(self):
        def test_fn(_):
            return returned_metrics
        db_man = DbManager(self.conf.json_db_path)
        db_man.delete_all()
        desc = """
        Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore 
        magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo 
        consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla pariatur. 
        Excepteur sint occaecat cupidatat non proident, sunt in culpa qui officia deserunt mollit anim id est laborum.
        """
        fn = tmt_recorder('test_exp', config_path='tests/test_config.json', description=desc)(test_fn)
        _ = fn(None)
        entry = context_manager.get().entry
        db_man = DbManager(context_manager.get().config.json_db_path)
        db_entry = db_man.get_entry_by_id(entry.id)
        self.assertEqual(db_entry.description, desc)




from tmt.storage.schema import *
from tmt.storage.json_db import DbManager
from tests import BaseTest
from datetime import datetime


@dataclass
class T1(BaseJsonDataclass):
    a: int
    b: typing.List[int]
    c: typing.Optional[typing.List[str]]
    d: typing.Optional[str]
    e: typing.List[typing.List[int]]


@dataclass
class T2(BaseJsonDataclass):
    t1: T1
    t2: typing.Optional[T1]


class TestSchema(BaseTest):

    def test_from_dict(self):
        d = {'a': 3, 'b': [1,2,3], 'c': ['x', 'y', 'z'], 'd': 'd', 'e': [[1,2], [3,4]]}
        d2 = {'a': 3, 'b': [1,2,3], 'c': None, 'd': None, 'e': [[1,2], [3,4]]}
        d3 = {'t1': d, 't2': d2}
        d4 = {'t1': d, 't2': None}
        t1 = T1.from_dict(d)
        for k, v in d.items():
            self.assertEqual(getattr(t1, k), v)
        t2 = T1.from_dict(d2)
        for k, v in d2.items():
            self.assertEqual(getattr(t2, k), v)

        t3 = T2.from_dict(d3)
        self.assertTrue(isinstance(t3.t1, T1))
        self.assertTrue(isinstance(t3.t2, T1))

        t4 = T2.from_dict(d4)
        self.assertTrue(isinstance(t4.t1, T1))
        self.assertIsNone(t4.t2)

    def test_from_json(self):
        import copy
        d = {'id': 0, 'name': 'test', 'date_created': int(datetime.now().timestamp()), 'local_results_path': '', 'metrics': [
            {'name': 'm', 'value': 0.0}, {'name': 'm1', 'value': 0.0}
        ], 'other_runs': [], 'results': []}
        Entry.from_json(json.dumps(d))
        d['other_runs'] = [copy.deepcopy(d)]
        Entry.from_json(json.dumps(d))

    def test_add_entry(self):
        db = DbManager(self.conf.json_db_path)
        entry = Entry(id='21jf10jf', name='asdf', args='asdf', date_created=Timestamp(0), local_results_path='')
        db.add_new_entries([entry])
        with open(db.db_path, 'r') as f:
            data = json.load(f)['data']
            for d in data:
                if d['id'] == entry.id:
                    return
        self.fail('Data was not added')

    def test_delete_entry(self):
        db = DbManager(self.conf.json_db_path)
        entry = Entry(id='21jf10jf', name='asdf', args='asdf', date_created=Timestamp(0), local_results_path='')
        db.add_new_entries([entry])
        self.assertTrue(db.delete_entry(entry))
        with open(db.db_path, 'r') as f:
            data = json.load(f)['data']
            for d in data:
                self.assertFalse(d['id'] == entry.id)

    def test_update_entry(self):
        db = DbManager(self.conf.json_db_path)
        entry = Entry(id='21jf10jf', name='asdf', args='asdf', date_created=Timestamp(0), local_results_path='')
        db.add_new_entries([entry])
        entry.name = 'updated'
        db.update_entries([entry])
        with open(db.db_path, 'r') as f:
            data = json.load(f)['data']
            d = next(filter(lambda d: d['id'] == entry.id, data))
            self.assertEqual(d['name'], entry.name)

    def test_search_by_regex(self):
        db = DbManager('tests/test_db_tui.json', read_only=True)
        self.assertGreater(len(db.get_entries_by_name_regex(r'test\d')), 0)



from tmt.interface.tui.main import MainApp


def main():
    import argparse
    parser = argparse.ArgumentParser('That Metric Timeline (TMT) terminal user interface (TUI).\nThe goal of this TUI '
                                     'is to provide an easy way for the user to look for experiments and look at some '
                                     'of their details. Operations like looking at the code snapshot backup, '
                                     'loading results and so on are left to the user (the TmtManager helper class is '
                                     'available for some of these operations).')
    parser.add_argument('--config', '-c', help='configuration path. If not given, TUI will try to load from the '
                                               'default configuration path or a default configuration')

    args = parser.parse_args()
    MainApp(args.config).run()


if __name__ == '__main__':
    main()



from __future__ import annotations
from tmt.storage.schema import Entry, Result
from tmt.storage.json_db import DbManager
from uuid import uuid4
from datetime import datetime
from contextvars import ContextVar
from tmt.configs.parser import Configs
from typing import Optional, Any, Callable
from tmt.exceptions import DuplicatedNameError
from tmt.utils.duplicates import DuplicateStrategy, DuplicatePolicy
import os
import sys
import pickle
import warnings

context_manager: ContextVar[ContextManager] = ContextVar('context_manager', default=None)


class ContextManager:
    def __init__(self, name: str, config_path: Optional[str] = None, duplicate_strategy=DuplicateStrategy(), description=""):
        if config_path:
            self.config = Configs.from_config(config_path)
        else:
            self.config = Configs.default_config()
        self.entry = Entry(
            id=str(uuid4()),
            name=name,
            description=description,
            args=' '.join(sys.argv),
            date_created=int(datetime.now().timestamp()),
            local_results_path=self.config.results_path,
        )
        self.duplicate_strat = duplicate_strategy
        self.parent = DbManager(self.config.json_db_path).get_entries_by_name(name)
        if self.parent:
            if self.duplicate_strat.policy is DuplicatePolicy.DONT_ALLOW:
                raise DuplicatedNameError(f'one (or more) entry with name {name} already exists. Set '
                                          f'`duplicate_strategy` or change name (recommended)')
            elif self.duplicate_strat.policy is DuplicatePolicy.AS_SUB_ENTRY:
                if len(self.parent) > 1:
                    parent_id = self.duplicate_strat.parent_id
                    parent_dict = {e.id: e for e in self.parent}
                    if not self.duplicate_strat.parent_id or self.duplicate_strat.parent_id not in parent_dict:
                        parent = sorted(self.parent, key=lambda e: e.date_created, reverse=True)[0]
                        warnings.warn(f'{len(self.parent)} entries already exist with this name and a '
                                      f'valid DuplicateStrategy.parent_id was not specified. I will use {parent.short_str()} as '
                                      f'the parent entry. See documentation for DuplicateStrategy.')
                        parent_id = parent.id
                    self.parent = parent_dict[parent_id]
                else:
                    self.parent = self.parent[0]
                self.parent.other_runs.append(self.entry)
            elif self.duplicate_strat.policy is DuplicatePolicy.AS_NEW_ENTRY:
                self.parent = None
        self.snap_manager = self.config.init_snapshot_manager(self.entry.id)
        self.entry.local_snapshot_path = self.snap_manager.snapshot_dest
        os.makedirs(self.get_save_path(), exist_ok=True)
        self.last_saved_counter = 0

    @property
    def root_entry(self) -> Entry:
        if self.parent and self.duplicate_strat.policy is DuplicatePolicy.AS_SUB_ENTRY:
            return self.parent
        return self.entry

    def save(self, obj: Any, name: str, allow_exist=False, extension='.pkl',
             custom_save: Optional[Callable[[Any, str], Optional[str]]] = None) -> str:
        path = self.get_save_path_with_name(name)
        if os.path.exists(path):
            if not allow_exist:
                raise DuplicatedNameError(f'{self.get_save_path_with_name(name)} already exists and `allow_exist` is '
                                          f'False. Specify another name to save this object')
            else:
                path = self.increment_last_saved_path(name)
        path += extension
        if custom_save is not None:
            custom_path = custom_save(obj, path)
            path = custom_path if custom_path is not None else path
        else:
            with open(path, 'wb') as f:
                pickle.dump(obj, f)
        self.entry.results.append(Result(self.entry.id, name, path))
        return path

    def get_save_path_with_name(self, name: str):
        return os.path.join(self.get_save_path(), name)

    def get_save_path(self):
        return os.path.join(self.entry.local_results_path, self.entry.id)

    def increment_last_saved_path(self, name: str):
        self.last_saved_counter += 1
        return os.path.join(self.get_save_path(), f'{name}_{self.last_saved_counter}')



import os

from tests import BaseTest
from tmt import tmt_recorder, tmt_save
from tmt.utils.manager import TmtManager
import numpy as np

from tmt.history.context import context_manager
from tmt.storage.json_db import DbManager

returned_metrics = {'f1': 0.87, 'acc': 0.45, 'loss': 1e-4}


def custom_save(obj, path):
    np.save(path, obj)


def custom_save_path(obj, _):
    path = 'tests/custom_save.npy'
    np.save('tests/custom_save.npy', obj)
    return path


@tmt_recorder('test_exp_custom_save', config_path='tests/test_config.json')
def decorated_fn(_):
    # do something
    tmt_save(returned_metrics, name='test_metrics', extension='.npy', custom_save=custom_save)
    return returned_metrics


@tmt_recorder('test_exp_custom_save_path', config_path='tests/test_config.json')
def decorated_fn_path(_):
    # do something
    arr = np.array([0])
    tmt_save(arr, name='test_metrics', extension='.npy', custom_save=custom_save_path)
    return returned_metrics


class TestUtils(BaseTest):

    def tearDown(self):
        super().tearDown()
        os.remove('tests/custom_save.npy')

    def test_custom_save(self):
        _ = decorated_fn(None)
        entry = context_manager.get().entry
        db_man = DbManager(context_manager.get().config.json_db_path)
        db_entry = db_man.get_entry_by_id(entry.id)
        self.assertEqual(len(db_entry.results), 1)
        self.assertEqual(db_entry.results[0].path[-4:], '.npy')
        np.load(db_entry.results[0].path, allow_pickle=True)
        _ = decorated_fn_path(None)
        entry = context_manager.get().entry
        db_man = DbManager(context_manager.get().config.json_db_path)
        db_entry = db_man.get_entry_by_id(entry.id)
        self.assertEqual(len(db_entry.results), 1)
        self.assertEqual(db_entry.results[0].path, 'tests/custom_save.npy')
        self.assertEqual(np.load(db_entry.results[0].path), np.array([0]))

        manager = TmtManager(config='tests/test_config.json')
        manager.set_entry_by_name("test_exp_custom_save_path")
        gen = manager.load_results()
        self.assertWarns(UserWarning, next, gen)



from tmt.history.snapshot import *
import unittest


class TestHistory(unittest.TestCase):
    def setUp(self):
        self.snap = SnapshotManager('test_id', 'tests', 'tests/snapshot_test', 'tests/fake_project', 'tests/snapshot_test/last', 'tests/.gitignore')
        self.new_folder = os.path.join(self.snap.snapshot_source, 'new_folder')

    def tearDown(self):
        if os.path.exists(self.snap.snapshot_target):
            shutil.rmtree(self.snap.snapshot_target)
        if os.path.exists(self.snap.last_snapshot_link):
            os.remove(self.snap.last_snapshot_link)
        if os.path.exists(self.new_folder):
            shutil.rmtree(self.new_folder)

    def test_make_first_snapshot(self):
        self.assertFalse(os.path.exists(self.snap.last_snapshot_link))
        self.snap.make_snapshot()
        self.__assert_no_diff()
        self.assertTrue(os.path.islink(self.snap.last_snapshot_link))
        self.assertEqual(os.readlink(self.snap.last_snapshot_link), self.snap.snapshot_dest)

    def test_make_snapshot_with_diff(self):
        if not os.path.exists(self.snap.last_snapshot_link):
            self.test_make_first_snapshot()
        diff_file = os.path.join(self.snap.snapshot_source, '__init__.py')
        with open(diff_file, 'a') as f:
            f.write('\nprint("test")')
        init = self.snap.__dict__
        init['id'] = 'test_id_1'
        del init['snapshot_dest']
        self.snap.make_snapshot()
        self.__assert_no_diff()
        self.assertTrue(os.readlink(self.snap.last_snapshot_link), self.snap.snapshot_dest)
        for root, _, files in os.walk(self.snap.snapshot_dest):
            for f in files:
                if f in self.snap.gitignored_files:
                    continue
                if os.path.join(root, f) == os.path.join(self.snap.snapshot_dest, '__init__.py'):
                    continue
                self.assertTrue(os.stat(os.path.join(root, f)).st_nlink > 1)
        with open(diff_file, 'w') as f:
            f.write('')

    def test_make_snapshot_with_new_folder(self):
        if not os.path.exists(self.snap.last_snapshot_link):
            self.test_make_first_snapshot()
        os.mkdir(self.new_folder)
        with open(os.path.join(self.new_folder, 'test.py'), 'w') as f:
            f.write('print("")')
        init = self.snap.__dict__
        init['id'] = 'test_id_1'
        del init['snapshot_dest']
        self.snap.make_snapshot()
        self.__assert_no_diff()

    def test_make_first_snapshot_without_ignore(self):
        self.snap.ignore_path = None
        self.test_make_first_snapshot()
        self.tearDown()
        self.snap.ignore_path = ''
        self.test_make_first_snapshot()
    
    def test_make_snapshot_with_diff_without_ignore(self):
        self.snap.ignore_path = None
        self.test_make_snapshot_with_diff()
        self.tearDown()
        self.setUp()
        self.snap.ignore_path = ''
        self.test_make_snapshot_with_diff()

    def __assert_no_diff(self):
        cmp = dircmp(self.snap.snapshot_source, self.snap.snapshot_dest)
        self.assertEqual(list(self.snap.get_diff_files(cmp, self.snap.snapshot_source)), [])
        return cmp



import enum
from dataclasses import dataclass


class DuplicatePolicy(enum.Enum):
    """
    Enum class used to be used with :py:class:`tmt.utils.duplicates.DuplicateStrategy`.
    Possible choices are:

    #. NO_ALLOW: do not allow duplicate names (i.e., throws :py:exc:`tmt.exceptions.DuplicatedNameError`);
    #. AS_SUB_ENTRY: allow duplicate names, insert this new one as a "sub-entry" of the pre-existing one;
    #. AS_NEW_ENTRY: allow duplicate names, create a completely new entry with the same name of the previous one
       (entries are still identified by a unique UUID).

    Having an entry as a "sub-entry" of another one means that it will be saved in the `entry.others` field.
    The sub-entry won't be directly searchable. Using a sub-entry is especially useful when running a task
    (e.g. experiment)
    which is somehow connected to that entry and still part of that previous task.

    .. note::
        When :py:meth:`tmt.utils.manager.TmtManager.load_results` is called on an entry with a sub-entry, results
        for the sub-entry will be returned as well.

    .. note::
        Specifying an entry as a sub-entry will still take a snapshot of the current code etc. Nothing changes in how
        the entry is saved and processed, except for what stated before.

    .. warning::
        If you specify :py:attr:`tmt.utils.duplicates.DuplicateStrategy.AS_SUB_ENTRY` and there are multiple entries
        with the same name already available, `tmt` cannot tell which one you would like to use. You can specify a
        `parent_id` when creating :py:class:`tmt.utils.duplicates.DuplicateStrategy`. Otherwise, the last created entry
        will be used.
    """
    DONT_ALLOW = enum.auto()
    AS_SUB_ENTRY = enum.auto()
    AS_NEW_ENTRY = enum.auto()


@dataclass
class DuplicateStrategy:
    """
    Data class used with :py:class:`tmt.decorators.recorder` to decide how to handle duplicates.
    `parent_id` is used when :py:class:`tmt.utils.duplicates.DuplicatePolicy.AS_SUB_ENTRY` is specified and more than one
    entry is found for a given name. In this case, if you don't specify a valid `parent_id`, a warning will be raised
    and the last created entry (with the given name) will be used as the parent.
    """
    policy: DuplicatePolicy = DuplicatePolicy.DONT_ALLOW
    parent_id: str = ""



import enum


class Versions(str, enum.Enum):
    ZERO_SEVENTEEN = "0.1.7"
    ZERO_EIGHTEEN = "0.1.8"

    @staticmethod
    def get_last() -> str:
        return Versions.ZERO_EIGHTEEN


__version__ = Versions.get_last()



from tmt.history.context import context_manager
from typing import Any, Optional, Callable


def save(obj: Any, name: str, allow_exist=False, extension='.pkl',
         custom_save: Optional[Callable[[Any, str], Optional[str]]] = None) -> str:
    """
    This function can be used to save any `pickable <https://docs.python.org/3/library/pickle.html#what-can-be-pickled-and-unpickled>`_. object you may want to save when running your experiments.
    It must be called by a function decorated with :py:func:`tmt.decorators.recorder.recorder` (or within code called from that function).

    .. note::
        For simplicity, ``tmt`` exposes this function from its root with the ``tmt_save`` name. You can import it with ``from tmt import 
        tmt_save``

    What this function does is not too far from simply doing:

    .. code-block:: python

        with open(library_path + name, 'wb') as f:
            pickle.dump(obj, f)
    
    Using this function however, we make sure that `tmt` knows where all of your pickled objects are so that you can 
    later easily retrieve them.

    **Usage**:

    .. code-block:: python

        from tmt import tmt_recorder, tmt_save

        @tmt_recorder(name="some_experiment_with_data")
        def train_and_predict(...):
            ...
            preds = lr.predict(x_te)
            tmt_save(preds, name='lr_predictions')
            return {'f1': f1_score(y_te, preds), 'accuracy': accuracy_score(y_te, preds)}

    Should you need it, you can pass a `custom_save` function, in order to use another saving function instead of pickle.
    This can be done also if you want to save the object in another path.

    **Example with custom saving function**:

    .. code-block:: python

        def my_save_fn(obj, path):
            np.save(path, obj)

        # ...
        def train_and_predict(...):
            # ...
            tmt_save(preds, name='lr_predictions', custom_save=my_save_fn, extension='.npy')

    :param obj: the object you want to save, must be pickable.
    :type obj: Any
    :param name: the object will be saved with this name. Use a name that will help you recognize this object in the future.
    :type name: str
    :param allow_exist: if True, allows this object to be saved even if the name already exists. Defaults to False.
    :type allow_exist: bool, optional
    :param extension: the extension to save the file with, defaults to '.pkl'.
    :type extension: str, optional
    :param custom_save: a user-defined function to save the object (e.g., you can use `numpy.save`). The function will
        receive the object to save and the file path. If you want to save to an alternative path, make sure to return this
        path.
    :type custom_save: Callable[[Any, str], Optional[str]], optional
    :raises ValueError: this is raised if the function is illegaly called outside a function decorated with :py:func:`tmt.decorators.recorder.recorder`
    :raises tmt.exceptions.DuplicatedNameError: if ``name`` already exists and ``allow_exist`` is False, this exception will be raised.
    :return: the path where the object was saved. ``tmt`` will take care of storing it, so you can ignore this return value.
    :rtype: str
    """
    context = context_manager.get()
    if context is None:
        raise ValueError("`save` function was called before initializing the context. "
                         "Did you use a tmt decorator/class?")
    return context.save(obj, name, allow_exist, extension, custom_save=custom_save)

